/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Annozilla.
 *
 * The Initial Developer of the Original Code is
 * Doug Daniels <rainking@alumni.rice.edu>
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   C. Greg Hagerty <cgreg@cgreg.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//primarily for debugging, but now also to notify about stylesheet
var gConsoleService = Components.classes['@mozilla.org/consoleservice;1']
    .getService(Components.interfaces.nsIConsoleService);

//function dumpToConsole(msg) {gConsoleService.logStringMessage (msg);}
function dumpToConsole(msg) {}

function msgToConsole(msg) {gConsoleService.logStringMessage (msg);}

/**
 * Return values for evaluateNode 
 */
// the node is before the highlighting; exclude it
const BEFORE_RANGE_EXCLUDE = 1; 
// the node is where highlighting begins (but not ends); exclude it
const BEGIN_RANGE_EXCLUDE = 2;
// the node is in the range of highlighting; include it
const IN_RANGE_INCLUDE = 3;
// the node is where the range of highlighting ends; include it
const END_RANGE_INCLUDE = 4;
// the node is where the range of highlighting ends; exclude it
const END_RANGE_EXCLUDE = 5;

/* NOTE:
   need to check whether the currentNode is the endContainer *before*
   we check whether it is the start container, so that we don't have the
   problem of changing currentNode in midstream and perhaps making
   currentNode != hLiteRange.startNode when it otherwise would be.  that
   is, going backward doesn't have the problems that going forward does,
   because splitText preserves the left-side node in the split. */

// Stylesheet to be added for highlighting, if one with the filename isn't there
const ANNOTATIONS_STYLESHEET = "chrome://annozilla/content/annotations.css";
var gBeforeRange;


/**
 * Test drive the highlighter on
 * the current mozilla selection.
 */
/*
function testHighlight() {
    var seln = window._content.getSelection();
    var hRange = document.createRange();
    hRange.setStart(seln.anchorNode, seln.anchorOffset);
    hRange.setEnd(seln.focusNode, seln.focusOffset);
    doHighlight(hRange, 'Comment');
}
*/

 
/**
 * Highlight a range. (Public function.)
 * @param hRange
 * @param type Annotation type
 * @param annotation
 * @return An array of nodes which were highlighted
 */
function doHighlight(hRange, type) {
    var annotationID;

    // optional third argument: specific annotation
    if ((arguments.length > 1) && arguments[2].annotationURI) {
        annotationID = arguments[2].annotationURI;
    }

    //"page" and "point" annotations get special handling
    if (hRange.startContainer.nodeName.toLowerCase() == "html") {
        // xpointer service looks for doc.body to determine whether to
        // casify tag names - if we wrap the body, xptrs may not resolve
        // besides, we don't really want to highlight everything, it's slow
        dump("(Page annotation)\n");
        var nodeToInsert = hRange.startContainer.ownerDocument.createElementNS(
             "http://www.w3.org/1999/xhtml",
             "span");
        xptrMarkElement(nodeToInsert);
        nodeToInsert.className='annotation';
        nodeToInsert.setAttribute('annotation-type', type);

        //Create and insert empty highlight span to allow for special footnote
        //Note: display attribute won't hide from CSS :before nor :after
        // nodeToInsert.setAttribute('display', 'none');
        // but CHANGING THE SEGMENT ATTRIBUTE MAY HIDE IT FROM CSS, IF DESIRED
        nodeToInsert.setAttribute('segment', 'first-last');

        if (annotationID) {
            nodeToInsert.setAttribute('annotation-id', annotationID);
        }
        hRange.startContainer.insertBefore(nodeToInsert, 
                                          hRange.startContainer.firstChild);
        return;
    } else if ((hRange.startContainer == hRange.endContainer)
               && (hRange.startOffset == hRange.endOffset)) {
        dump("Point annotation\n");
        //proceed, but might want to make span hidden
    }

    var rangeList = highlightRange(hRange);
/*
    dump("Highlighting: \""+hRange.toString()+"\"\n");
    // enumerate rangeList
   dumpToConsole("enumerating rangeList:\n");
   for (var i=0; i < rangeList.length; i++) {
       dumpToConsole(" rangeList[" + i + "]==" + rangeList[i].toString() + 
            " with startContainer==" + rangeList[i].startContainer + 
            ", startOffset==" + rangeList[i].startOffset + 
            " with endContainer==" + rangeList[i].endContainer + 
            ", endOffset==" + rangeList[i].endOffset + "\n");
   }
*/
    var console = Components.classes['@mozilla.org/consoleservice;1']
        .getService(Components.interfaces.nsIConsoleService);

    for (var i = 0; i < rangeList.length; i++) {

        console.logStringMessage (i + "] " + rangeList[i]);

        var ref_node = rangeList[i].startContainer.wrappedJSObject;
        console.logStringMessage (i + "] " + ref_node);
        try {
            ref_node.setAttribute ("style", "-moz-binding: url('chrome://annozilla/content/bindings/annozillaBindings.xml#highlight')");
        } catch (e) {
            console.logStringMessage(e);
        }
    }
    return;
}
                           
/* Add a stylesheet link to the document 
 *   if it - or one with the same filename - isn't there already
 *  presumably someone could then override the default cssURI
 *  a better way would be to allow a global annotation to override it
 *  TODO: Explain how to provide custom stylesheets in the documentation
 *
 * @param document the document to add stylesheet link to
 * @param cssURI the stylesheet URI
 */
function addStyleSheet(document, cssURI) {
    var stylesheets=document.styleSheets;
    var found=false;
    if (stylesheets) {
        var cssName = cssURI;
        var shorterName = cssURI.substr(cssURI.lastIndexOf("/")+1);
        if (shorterName.length > 0) {cssName = shorterName;}
        var namelen=cssName.length;
        var neglen=-1*namelen;
        for (var i = 0; i < stylesheets.length; i++) {
            if (stylesheets[i].href.substr(neglen, namelen) == cssName) {
                found=true; 
            }
        }  
    }
    if (!found) { 
        //msgToConsole("Stylesheet "+cssName+" not found: using "+cssURI);
        //dump("Stylesheet "+cssName+" not found: using "+cssURI+"\n");
        var headElementList = document.getElementsByTagName("head");
        var linkElement = document.createElement("link");
        linkElement.setAttribute("href", cssURI);
        //  linkElement.setAttribute("id", id);
        linkElement.setAttribute("rel", "stylesheet");
        linkElement.setAttribute("type", "text/css");               
        if (headElementList.length > 0) {
            headElementList[0].appendChild(linkElement);
        } else {
            document.documentElement.appendChild(linkElement);
        }
    }
}


/**
 * Calculates a list of range
 * that exactly covers the input range and all
 * have their startContainer == endContainer, and
 * thus may be surrounded by a span or similar tag.
 *
 * @param range nsIDOMRange to cover
 * @return Array of Range
 */
function highlightRange(range) {
    /* ensure that range is valid (not collapsed, has start and end containers, start
       and end offsets) */
    // FIXME -- todo

    range.startContainer.ownerDocument.loadBindingDocument
          ("chrome://annozilla/content/bindings/annozillaBindings.xml");

    gBeforeRange = true;

    var rangeList = [];
    dumpToConsole("highlight range: " + range.toString() 
         + " start: " + range.startContainer.nodeName
         + "+" + range.startOffset
         + " end: " + range.endContainer.nodeName
         + "+" + range.endOffset + "\n");
    var ancestorContainer = range.commonAncestorContainer;
    // ancestorContainer must be an element to ensure that splitting text nodes will work
    if (ancestorContainer.childNodes.length == 0) {
       ancestorContainer = ancestorContainer.parentNode;
    }
    var topLevelRetVal = evaluateNode(ancestorContainer, range, rangeList);
    if (topLevelRetVal == END_RANGE_INCLUDE) {
       var newRange = document.createRange();
       newRange.selectNode(range.commonAncestorContainer);
       rangeList.push(newRange);
    }
    // do nothing
    else if (topLevelRetVal == END_RANGE_EXCLUDE) {}
    else { 
        throw new Error("highlightRange: got return value " + topLevelRetVal + " in error");
    }
        
    return rangeList;
}

/**
 * Evaluates a Node to determine whether it is completely covered by a range.
 * 
 * @param currentNode DOM Node to evaluate
 * @param range nsIDOMRange of highlighted selection
 * @param rangeList accumulated array of ranges to cover the highlighted selection
 * @return how to handle the node -- one of codes at top of file
 */
function evaluateNode(currentNode, hLiteRange, rangeList) {
    var newRange = null;
    var isStartContainer = false;
    var isEndContainer = false;
    
    var currentNodeIsCharData = isCharDataNode(currentNode);
    dumpToConsole("evaluateNode called for node: " + currentNode.nodeName + 
                  " for which charNodeType==" + currentNodeIsCharData + " and rangeList.length==" + rangeList.length + "\n");
    // if the currentNode is the endContainer of the highlighted range
    if (hLiteRange.endContainer == currentNode) {
        isEndContainer = true;
        
        dumpToConsole("found endContainer: " + hLiteRange.endContainer.nodeName + "\n");
        
        // CHARACTER DATA type node
        if (currentNodeIsCharData) {
            
            /* if the highlighted range does not extend to the end of the text node, 
               split the text node at the endOffset to get the portion of the text
               node that actually *is* covered. afterward, the currentNode will exactly
               cover the end of the highlighted selection */
            if (hLiteRange.endOffset < currentNode.length) {
                currentNode.splitText(hLiteRange.endOffset);
            }   
            // otherwise, do nothing.  currentNode is already correct.
            else {}
            
        }
        // non-CHARACTER DATA type node (Element)
        else {
            // if the endOffset isn't the last child in the node
            if (hLiteRange.endOffset < currentNode.childNodes.length) {
                newRange = document.createRange();
                // set the start to be before the first child of currentNode
                newRange.setStart(currentNode, 0);
                // set end to be endOffset
                newRange.setEnd(currentNode, hLiteRange.endOffset);
            }
            /* if endOffset covers all of currentNode, do nothing */
            else {}
        }
    }
    
    /* if current node is the startContainer of the highlighted
       range.  It may *also* have been the endContainer above */
    if (hLiteRange.startContainer == currentNode) {
        
        dumpToConsole("found startContainer: " + hLiteRange.startContainer.nodeName + "\n");
        
        isStartContainer = true;
        
        // start-node is a CHARACTER DATA type Node
        if (currentNodeIsCharData) {
            
            // if the highlighted text does not begin at offset 0
            if (hLiteRange.startOffset > 0) {
                /* we must now split the currentNode to capture
                   only what has been highlighted.  this will mess
                   up the highlighted range hLiteRange, since it should actually
                   start at the beginning of the new node we're creating
                   through the split.  we will have to update hLiteRange */
                var newStartNode = currentNode.splitText(hLiteRange.startOffset);
                // first, reset the end if it has been split -- this can
                // only happen if the start and end containers are equal
                if (hLiteRange.endContainer == currentNode) {
                    hLiteRange.setEnd(newStartNode, newStartNode.nodeValue.length);
                }
                // reset the start
                hLiteRange.setStart(newStartNode, 0);
                
                /* return a before range exclude -- the next child processed by the
                   parent will now be the start container */
                return BEFORE_RANGE_EXCLUDE;
            }
            /* otherwise, if the highlighted text begins at 0, do nothing,
               as we may be able to include this node in a higher level 
               if it is completely covered */
            else {}
        } // end if CHARACTER DATA type node
        
        // start-node is non-CHARACTER DATA (Element) type Node
        else {
            // if highlighted range starts after the first child
            if (hLiteRange.startOffset > 0) {
                /* if there is a newRange already, it means that
                   we've already set a valid ending point for this range.
                   as such, we only need set the correct beginning point */
                if (newRange) {
                    newRange.setStart(currentNode, hLiteRange.startOffset);
                }
                /* otherwise, no valid ending point yet, so 
                   we must create the range and extend it to the end of
                   the currentNode's children */
                else {
                    newRange = document.createRange();
                    newRange.setStart(currentNode, hLiteRange.startOffset);
                    newRange.setEnd(currentNode, currentNode.childNodes.length);
                }
            }
            /* otherwise, if the startOffset is before the
               first child of currentNode, so we don't need to adjust anything */
            else {}
        } // end else NON-CHARACTER DATA type NODE
        // set the global variable saying that we're
        // no longer before the highlighted range
        // this must be set here, and not at the top,
        // because there is one case above that splits
        // a text node and is thus still before the range
        gBeforeRange = false;
        
    } // end if currentNode is the highlighted range startContainer
    
    // if there's a newRange, add it to the list
    if (newRange) {
        rangeList.push(newRange);
    }
    
    // if endContainer was found
    if (isEndContainer) {
        if (newRange) {
            return END_RANGE_EXCLUDE;
        }
        else {
            return END_RANGE_INCLUDE;
        }
    }
    
    // if ! endContainer && startContainer
    if (isStartContainer) {
        if (newRange) {
            return BEGIN_RANGE_EXCLUDE;
        }
        else {
            return IN_RANGE_INCLUDE;
        }
    }
    
    /* otherwise, we are not at a boundary point
       and are inside of the highlighted range */
    
    // if we're at a TEXT Node 
    if (currentNodeIsCharData) {
        // if we're still before the range begins, return such
        if (gBeforeRange) {
            return BEFORE_RANGE_EXCLUDE;
        }
        // otherwise, include us
        else {
            return IN_RANGE_INCLUDE;
        }
    }
    
    /* otherwise, we're at a non-Tex node. 
       need to visit the children in order */
    
    /*  offset to begin including in the range -- 0 is before first child*/
    var includeBeginOffset = 0;
    /* offset to finish including in the range -- length is after last child*/
    var includeEndOffset = -1;
    // whether we see the end of the range in one of our children
    var endRangeFound = false;
    var endRangeIncluded = false;
    
    for (var i = 0; i < currentNode.childNodes.length; i++) {
        dumpToConsole("calling evaluateNode for child[" + i + "] of " + currentNode.nodeName + "\n");
        var retVal = evaluateNode(currentNode.childNodes.item(i), hLiteRange, rangeList);
        dumpToConsole("child[" + i + "] of " + currentNode.nodeName + " returned " + retValStringFromInt(retVal) + "\n");
        if (retVal == BEFORE_RANGE_EXCLUDE) {
            includeBeginOffset += 1;
            if (includeBeginOffset != (i + 1)) {
                //we may also be INSIDE? let's be forgiving
                includeBeginOffset -= 1;
                //throw new Error("invalid logic--includeBeginOffset was not == i+1, meaning " +
                dump("Warning:invalid logic--includeBeginOffset was not == i+1, meaning " +
                     "that a beforeRangeExclude was thrown incorrectly at time " + i);
            }
        }
        else if (retVal == BEGIN_RANGE_EXCLUDE) {
            includeBeginOffset += 1;
            if (includeBeginOffset != (i + 1)) {
                //we may also be INSIDE? let's be forgiving
                includeBeginOffset -= 1;
                //throw new Error("invalid logic--includeBeginOffset was not == i+1, meaning " +
                dump("Warning:invalid logic--includeBeginOffset was not == i+1, meaning " +
                     "that a beginRangeExclude was thrown incorrectly at time " + i);
            }
        }
        else if (retVal ==  IN_RANGE_INCLUDE) { }
        else if (retVal == END_RANGE_INCLUDE) {
            includeEndOffset = i + 1;
            // end the for loop 
            endRangeFound = true;
            endRangeIncluded = true;
            break;
        }
        else if (retVal == END_RANGE_EXCLUDE) {
            includeEndOffset = i;
            // end the for loop 
            endRangeFound = true;
            break;
        }
        else {
            throw new Error("Unexpected case: " + retVal + " in evaluation of child[ " + i + "] == " + 
                            currentNode.childNodes.item(i).nodeName);
        }
    } // end for loop over children
    
    /* if includeEndOffset is still -1, it means
       we didn't find the end offset anywhere within
       our descendants, and thus should include to the last one.
       we have to set includeEndOffset correctly here, because
       in the process of visiting all children, a split text node
       may cause the length to increase */
    if (includeEndOffset == -1) {
        includeEndOffset = currentNode.childNodes.length;
    }
    
    // if no children are included, 
    // we are either before the range or at the end
    if (includeBeginOffset == includeEndOffset) {
        // at then end of the range, not including
        // any children fully
        if (endRangeFound) { 
            return END_RANGE_EXCLUDE;
        }
        // before the range completely
        else {
            if ((currentNode.childNodes.length==0) && !gBeforeRange) {
                return IN_RANGE_INCLUDE;
            }
            return BEFORE_RANGE_EXCLUDE;
        }
    }
    
    // if only some children included
    if (includeBeginOffset != 0 ||
        includeEndOffset != currentNode.childNodes.length) {
        // put these children in a range
        newRange = document.createRange();
        newRange.setStart(currentNode, includeBeginOffset);
        newRange.setEnd(currentNode, includeEndOffset);
        
        // add the range
        //rangeList.push(newRange);
        //This is tricky! Try to maintain order of appearance:

        if(endRangeFound) {
            if (endRangeIncluded) {
                rangeList.push(newRange);
            } else {
                //end range must have just been pushed and newRange belongs just before it
                rangeList.splice(rangeList.length-1, 0, newRange);
            }
            return END_RANGE_EXCLUDE;
        }
        else {
            rangeList.push(newRange);
            return BEGIN_RANGE_EXCLUDE;
        }
    }
    
    // otherwise, all children included
    if (endRangeFound) {
        return END_RANGE_INCLUDE;
    }
    else {
        return IN_RANGE_INCLUDE;
    }
}


/**
 * Is this node a character-data or text type?
 */
function isCharDataNode(node) {
    return ((node.nodeType == node.TEXT_NODE) ||
            (node.nodeType == node.CDATA_SECTION_NODE));
}


/**
 * A version of the Range method surround contents
 * that works in the limited context we need it.
 * Assumes that the range has startContainer == endContainer
 * (which ours always do) and has non-Text node startContainer
 * and endContainer.
 * Surrounds such ranges with the input tag.
 *
 * @param tag nsIDOMNode to surround range with
 * @param range nsIDOMRange to be surrounded
 */
function simpleSurroundContents(tag, range) {
    // if the startContainer or endContainer are not
    // the same element, scream
    if (range.startContainer != range.endContainer) {
        throw new Error("startContainer and endContainer not equal");
    }
    if ( (range.startContainer.nodeType == range.startContainer.TEXT_NODE) ||
         (range.endContainer.nodeType == range.endContainer.TEXT_NODE) ) {
        throw new Error("startContainer and/or endContainer are text nodes");
    }

    var parent = range.startContainer;

    // insert the tag before the start of the range
    var nodesToMove = [];
    for (var i = range.startOffset; i < range.endOffset; i++) {
        var node = range.startContainer.childNodes.item(i);
        nodesToMove.push(node);
    }

    parent.insertBefore(
        tag,
        range.startContainer.childNodes.item(range.startOffset));

    for (var i = 0; i < nodesToMove.length; i++) {
        tag.appendChild(parent.removeChild(nodesToMove[i]));
    }
}
        
/**
 * A utility function to turn the
 * return values from evaluateNode
 * into human-readable strings.
 *
 * @param code evaluateNode return value
 * @return string version
 */
function retValStringFromInt(code) {
    switch (code) {
        case BEFORE_RANGE_EXCLUDE:
            return "BEFORE_RANGE_EXCLUDE";
        case BEGIN_RANGE_EXCLUDE:
            return "BEGIN_RANGE_EXCLUDE";
        case IN_RANGE_INCLUDE:
            return "IN_RANGE_INCLUDE";
        case END_RANGE_INCLUDE:
            return "END_RANGE_INCLUDE";
        case END_RANGE_EXCLUDE:
            return "END_RANGE_EXCLUDE";
        default:
            throw new Error("Invalid code " + code);
    }
}

