/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Annozilla.
 *
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  C. Greg Hagerty <cgreg@cgreg.com>
 *   dump to console for debug
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * XPointerService interaction.
 */
const XPOINTER_SERVICE_CONTRACTID = "@mozilla.org/xpointer-service;1";
const ANNOT_NS = "http://www.w3.org/2000/10/annotation-ns#";

var gXPointerService = null;

if (Components.classes[XPOINTER_SERVICE_CONTRACTID]) {
    gXPointerService = Components.classes[XPOINTER_SERVICE_CONTRACTID].getService().QueryInterface(Components.interfaces.nsIXPointerService);

}

/* for debug */
var gConsoleService = Components.classes['@mozilla.org/consoleservice;1']
    .getService(Components.interfaces.nsIConsoleService);
function dumpToConsole(msg) {gConsoleService.logStringMessage (msg);}

//function dumpToConsole(msg) {}

function xptrIsServiceLoaded() { return (gXPointerService != null); }

function xptrResolveFragment(fragment, doc){
    return gXPointerService.parseXPointerToRange(fragment, doc);
}

function xptrResolveAnnotation(annotation, doc) {
    var range = null; // resolved range for xpointer
    // context is full uri including xpointer
    var context = annotation.data.getProperty(ANNOT_NS, "context");
    
    if (context) {
	// strip uri down to xpointer
	var index = context.indexOf("#");
	context = context.substring(index + 1);
    // Little hack. FF3 won't allow us to insert data at the root of the document.
    if (context=="xpointer(/html[1])") { context = "xpointer(/html[1]/body[1])" };
	if (index != -1) {
	    try {
		return xptrResolveFragment(context, doc);
	    }
	    catch(e) {
		dump("\nCouldn't resolve annotation " + annotation.annotationURI + ": " + e + "\n");
                dump("context: "+context+"\n");
                dumpToConsole("Couldn't resolve annotation " 
                                + annotation.annotationURI + ": " + e + "\n"
                                + "in context: "+context);
	    }
	}
	else {
	    dump("couldn't find # in context: " + context + "\n");
	}
    }
    else {
	dump("couldn't get context for annotation: " + annotation.annotationURI + "\n");
    }
    
    // couldn't resolve annotation
    return null;
}

function xptrCreateXPointer(seln, doc) {
    return gXPointerService.createXPointerFromSelection(seln, doc);
}

function xptrMarkElement(element) {
    gXPointerService.markElement(element);
}
